/***********************************************************************************************/
/************************************ main.c ***************************************************/
/***********************************************************************************************/

#define __MAIN
#include "mydefines.h"

/***********************************************************************************************/
/************************ Main function that sets up and spawns threads ************************/
/***********************************************************************************************/
int main(){

	/* control thread */
	pthread_t            control_thread_id = -1;
	pthread_attr_t       control_thread_attr;

	/* data recording thread */
	pthread_t            data_thread_id = -1;
	pthread_attr_t       data_thread_attr;

	/* gui thread */
	pthread_t            gui_thread_id = -1;
	pthread_attr_t       gui_thread_attr;


	/*---------------------------------------------------------------------------------------------*/
	/*-------------------------------- QNX I/O Initialization -------------------------------------*/
	/*---------------------------------------------------------------------------------------------*/

	/* qnx call to get access to I/O ports, must be logged on as root */
	mmap_device_io(32,0x300);  
	ThreadCtl(_NTO_TCTL_IO,NULL);

	/*---------------------------------------------------------------------------------------------*/
	/*---------------------------------- Board Initialization -------------------------------------*/
	/*---------------------------------------------------------------------------------------------*/

	/* initialize servotogo board */
	stgInitialize();

	/*---------------------------------------------------------------------------------------------*/
	/*----------------------------------- Message Que Setup ---------------------------------------*/
	/*---------------------------------------------------------------------------------------------*/
	
   /* set all message cues to null */
	InitMsg();

	/*---------------------------------------------------------------------------------------------*/
	/*---------------------------------------- Gui Setup ------------------------------------------*/
	/*---------------------------------------------------------------------------------------------*/

   /* set up terminal as a ncurses window, with color */
   MainWindow = initscr();
   if(has_colors()== TRUE){
		start_color();
      /* initialize color schemes - use high numbers so as not to ruin other pairs, three parameters are: number, foreground, background */
		init_pair(WB,COLOR_WHITE,COLOR_BLACK);				/* white */
		init_pair(YB,COLOR_YELLOW,COLOR_BLACK);			/* yellow */
		init_pair(RB,COLOR_RED,COLOR_BLACK);				/* red */
		init_pair(MB,COLOR_MAGENTA,COLOR_BLACK);			/* purple */
		init_pair(GB,COLOR_GREEN,COLOR_BLACK);				/* green */
		init_pair(BB,COLOR_BLUE,COLOR_BLACK);				/* blue */
		init_pair(CB,COLOR_CYAN,COLOR_BLACK);				/* cyan */
		init_pair(KB,COLOR_BLACK,COLOR_BLACK);				/* black */
   }
   else{
		PrintErrorMsg("No color support!");
   }

   werase(MainWindow);													/* clear the MainWindow */
	mvwin(MainWindow,2,2);                                   /* position window */
   nonl();																	/* disable newline translation */
   cbreak();																/* set input mode for this terminal to cbreak mode */
   noecho();																/* don't echo input */
   nodelay(MainWindow,TRUE);											/* set input/output to non-blocking */
   curs_set(1);															/* hide cursor */
   getmaxyx(MainWindow,maxy,maxx);									/* grab the dimensions of the terminal */
	keypad(stdscr,TRUE);													/* enable arrow keys, etc. */

	/* paint static labels on main screen */
   InitializeDisplay();
	DrawStaticInfo();

	/*---------------------------------------------------------------------------------------------*/
	/*---------------------------------- Spawn Curses Gui Thread ----------------------------------*/
	/*---------------------------------------------------------------------------------------------*/

	/* start gui thread - the thread that runs the curses user interface */
	if(pthread_attr_init(&gui_thread_attr) != EOK)
	   PrintErrorMsg("Couldn't initialize thread attribute: in gui");
	else {
		SchedGet(0,0,&gui_thread_attr.param);  /* pid and tid set to zero mean "this thread" */
		gui_thread_attr.param.sched_priority = 43; 
		gui_thread_attr.policy = SCHED_FIFO;
		gui_thread_attr.flags |= PTHREAD_EXPLICIT_SCHED;
		gui_thread_id = ThreadCreate(0,startGuiThread,NULL,&gui_thread_attr);
	}
	if (gui_thread_id == -1) {
		PrintErrorMsg("Couldn't start gui thread!");
		exit(-1);
	}
	delay(100);

	/*---------------------------------------------------------------------------------------------*/
	/*------------------------------------ Data Recording Setup -----------------------------------*/
	/*---------------------------------------------------------------------------------------------*/

	/* initialize a data storage cache and semaphore */
	sem_init(&Data_Cache_Sem,0,0);
	mallopt(MALLOC_FILLAREA, (int)1); /* turn on automatic area boundary checking - warning not generated until free() */
	mallopt(MALLOC_CKACCESS, (int)1);
	mallopt(MALLOC_CKCHAIN, (int)1);  /*very expensive computationaly - should eliminate */
	Data_Cache = (float *)malloc(sizeof(float)*RECORD_SIZE*NUMBER_OF_RECORDS);

	/*---------------------------------------------------------------------------------------------*/
	/*------------------------------ Spawn Data Recording Thread ----------------------------------*/
	/*---------------------------------------------------------------------------------------------*/
	
	/* start data recording thread - the thread that runs all data transfer to file */
	if(pthread_attr_init(&data_thread_attr) != EOK)
	   PrintErrorMsg("Couldn't initialize thread attribute: in data");
	else {
		SchedGet(0,0,&data_thread_attr.param);  /* pid and tid set to zero mean "this thread" */
		data_thread_attr.param.sched_priority = 33; 
		data_thread_attr.policy = SCHED_FIFO;
		data_thread_attr.flags |= PTHREAD_EXPLICIT_SCHED;
		data_thread_id = ThreadCreate(0,startDataThread,NULL,&data_thread_attr);
	}
	if (data_thread_id == -1){
	   PrintErrorMsg("Couldn't start record thread!");
		exit(-1);
	}
	delay(100);

	/*---------------------------------------------------------------------------------------------*/
	/*------------------------------------- Spawn Control Thread ----------------------------------*/
	/*---------------------------------------------------------------------------------------------*/

	/* start control thread - the thread that runs the control loop */
	if(pthread_attr_init(&control_thread_attr) != EOK)
	   PrintErrorMsg("Couldn't initialize thread attribute: in control");
	else {
		SchedGet(0,0,&control_thread_attr.param);  /* pid and tid set to zero mean "this thread" */
		control_thread_attr.param.sched_priority = 63; 
		control_thread_attr.policy = SCHED_FIFO;
		control_thread_attr.flags |= PTHREAD_EXPLICIT_SCHED;
		control_thread_id = ThreadCreate(0,startControlThread,NULL,&control_thread_attr);
	}
	if (control_thread_id == -1){
	   PrintErrorMsg("Couldn't start control thread!");
		exit(-1);
	}
	delay(100);

	/*---------------------------------------------------------------------------------------------*/
	/*------------------------------------- Exit Startup Thread -----------------------------------*/
	/*---------------------------------------------------------------------------------------------*/
   ThreadDestroy(0,-1,NULL);
   return(0);
}


/***********************************************************************************************/
/************************************ Control Thread *******************************************/
/***********************************************************************************************/
void *startControlThread(void *){

	int i;

	/* setup timing of thread */
	memset(&Interrupt_Event, 0, sizeof(struct sigevent));
	SIGEV_INTR_INIT(&Interrupt_Event);
   id = InterruptAttachEvent(_NTO_INTR_CLASS_EXTERNAL | IRQ_NUMBER, &Interrupt_Event, (_NTO_INTR_FLAGS_END | _NTO_INTR_FLAGS_TRK_MSK));

   /* temporarily disable interrupts from the board and setup freqeuncy of interrupts */
   out8(stg_Base_Address + CNTRL1, 0x08);
   InterruptFrequency = stgConfigureInterruptTimer(INT_FREQ);
	DT = 1/InterruptFrequency;
	DT_INV = InterruptFrequency;

   /* enable Servo-to-Go interrupts on T0  - clears any watchdog events */
   out8(stg_Base_Address + CNTRL1, 0x09);  /* 0x69 */
   cntrl1 = in8(stg_Base_Address + CNTRL1);
   if(cntrl1 & 0x80){
      PrintErrorMsg("watchdog high before start");
   }

	/* initialize filters now that DT has been established */
	InitLowpass(&axis_01_vel_filter, DT, 2, AXIS_01_FILTER_CUTOFF, AXIS_01_FILTER_ZETA);
	InitLowpass(&chan_01_vel_filter, DT, 2, CHAN_01_FILTER_CUTOFF, CHAN_01_FILTER_ZETA);

	/* set digital IO directions */
	stgSetDioDirABC(A_INPUT|B_INPUT|C_HI_OUTPUT|C_LOW_OUTPUT);
	stgSetDioDirD(D_OUTPUT);	

	/* initialize encoder parameters */
	for(i=0;i<8;i++) stgSetEncoderCounts(i,0);

	/* turn off all indexpulses */
	stgSetIDLEN(0x00); 
	
	/* allow gui to start now that everything is set up */
	SendMsg(TO_DISPLAY_MSG, START_GUI);

	/* ALWAYS clear the watchdog source in CNTRL1 by zeroing out the INT-T0 bit */
	out8(stg_Base_Address + CNTRL1,0x09);

   /* initialize cpu timer values */
   start = 0;

	/*---------------------------------------------------------------------------------------------*/
	/*---------------------------- Repeating Control Thread While Loop ----------------------------*/
	/*---------------------------------------------------------------------------------------------*/
   while (emsg != TERMINATE_CONTROL_THREAD) {

		/* wait for an interrupt to occur */
		if(InterruptWait_r( 0, NULL ) < 0){
			PrintErrorMsg("wait() timeout");
		}
		
		/* bookkeeping of time */
		old_start = start;
		start = ClockCycles();
		sample_time = start - old_start;

		/* check to see if there was a watchdog event */
		cntrl1 = in8(stg_Base_Address + CNTRL1);
		if(cntrl1 & 0x80) PrintErrorMsg("watchdog event");

		/* ALWAYS clear the watchdog source in CNTRL1 by zeroing out the INT-T0 bit */
		out8(stg_Base_Address + CNTRL1,0x09);
		
		/* interrupt wait() automaticall perform a mask, make sure we can see it again */
		InterruptUnmask((_NTO_INTR_CLASS_EXTERNAL | IRQ_NUMBER),id);
		
		/* increment heartbeat */
		beat += 1;
		
		/* main structure that executes control code */
		control_status_switch_structure();

		/* bookkeeping of time */
		stop = ClockCycles();
		elapsed = stop-start;

		/* record data to the cache */
		if(beat%FILE_IO_PERIOD == 0){
			if((Wait_for_RecordThread == TRUE) && (recording == TRUE)) buffer_data();
		}

		/* check and write to disk if buffer is full (even if "recording == FALSE") */
		write_buffer();
		
		/* check for and act on messages to control thread */
		get_control_messages();

	}
	/*---------------------------------------------------------------------------------------------*/
	/*----------------------- End of Repeating Control Thread While Loop --------------------------*/
	/*---------------------------------------------------------------------------------------------*/

   PrintErrorMsg("TERMINATING Control Thread...");
   stgStopInterrupts();

   /* you should zero your analog outputs and disable any amplifier digital enables here */

   InterruptDetach(id);

	stop_data_thread();

	/* terminate the gui */
	terminate_gui = TRUE;

   delay(100); //allow time for final draw of gui to complete

   free(Data_Cache);

	/* close out the control thread */
   ThreadDestroy(0,-1,NULL);
   return(0);
}


/***********************************************************************************************/
/*************************************** Data Thread *******************************************/
/***********************************************************************************************/
void *startDataThread(void *)
{
   /* This thread polls on a global cache of data and meters it to the file system  - 
    * at a priority lower than the control thread */
   
   unsigned int msg = EMPTY;
   FILE *fd_out = 0;
   int success = TRUE;
   char tmpstr2[80];
   time_t time_of_day;
   int first_time = TRUE;

   /* repeating while loop is located here ********************************************************/ 
	while((success == TRUE) && (terminate_data == FALSE)){
		/* wait to get awakened by semaphore */
      while(sem_wait(&Data_Cache_Sem) == EINTR);
      /* check for thread message, STOP, RENAME, RESUME */

			if((msg = GetMsg(TO_RECORD_MSG)) != EMPTY){
				switch(msg){
					case SUSPEND_RECORDING:
						PrintErrorMsg("USER pause recording...");
						recording = FALSE;
					break;
               case RESUME_RECORDING:
                  if (first_time == FALSE){
                     if(recording == FALSE){
                        PrintErrorMsg("USER resume recording...");
                        recording = TRUE;
                        Data_Cache_Index_Out = Data_Cache_Index_In;
                     }
                  } else {
                     /* open file for recording - name is the timeofday */
                     PrintErrorMsg("USER start recording...");
                     time_of_day = time( NULL );
                     strftime( filestr, 40, "./archive/%j-%H-%M-%S.dat",localtime( &time_of_day ) );
                     fd_out = fopen(filestr,"w");
                     if (fd_out == NULL) {
                        sprintf(tmpstr2,"Couldn't open %s: %d",filestr,errno);
                        PrintErrorMsg(tmpstr2);
                        success = FALSE;
                     }
                     first_time = FALSE;
                     recording = TRUE;
                     Data_Cache_Index_Out = Data_Cache_Index_In;
                  }
               break;
               case START_NEW_FILE:
                  PrintErrorMsg("USER start new file...");
                  fclose(fd_out);
                  time_of_day = time( NULL );
                  strftime( filestr, 40, "./archive/%j-%H-%M-%S.dat",localtime( &time_of_day ) );
                  fd_out = fopen(filestr,"w");
                  if (fd_out == NULL) {
                     sprintf(tmpstr2,"Couldn't open %s: %d",filestr,errno);
                     PrintErrorMsg(tmpstr2);
                     success = FALSE;
                  }
                  sprintf(filesizestr,"% 7.2f:(MB)",((double)filesize(fd_out))/1000.0/1000.0);
               break;
               case STOP_RECORDING:
                  PrintErrorMsg("User stop recording...");
                  recording = FALSE;
                  terminate_data = TRUE;
               break;/* end switch case */
            }
			}

			/* at this point, there's at least one record of data available to write - send one out and move on.. */
			if((recording == TRUE) && (Data_Cache_Index_In != Data_Cache_Index_Out) && (first_time == FALSE)){
				fwrite((float*)(Data_Cache + Data_Cache_Index_Out*RECORD_SIZE),sizeof(float),RECORD_SIZE,fd_out);
				if (++Data_Cache_Index_Out == NUMBER_OF_RECORDS) Data_Cache_Index_Out = 0;
				sprintf(filesizestr,"% 7.2f:(MB)",((double)filesize(fd_out))/1000.0/1000.0);
			}
	}
   /*******************************************************************************************************/ 

   if(success == TRUE){
      if(recording == TRUE){
			/* send out any remaining records ?  could just be making garbage... */
			while(Data_Cache_Index_In != Data_Cache_Index_Out){
				fwrite((float*)(Data_Cache + Data_Cache_Index_Out*RECORD_SIZE),sizeof(float),RECORD_SIZE,fd_out);
				if (++Data_Cache_Index_Out == NUMBER_OF_RECORDS) Data_Cache_Index_Out = 0;
			}
		}
      fclose(fd_out);
      PrintErrorMsg("Succesful data session.");
   }

   PrintErrorMsg("TERMINATING RecordThread...");

   /* semaphore destroyed and memory is freed in the control task */
   sem_destroy(&Data_Cache_Sem);
   
   /* send handshake back to control thread */
   SendMsg(TO_CONTROL_MSG,RELEASE_RECORDING);

   ThreadDestroy(0,-1,NULL);
   return(0);
   
}


/***********************************************************************************************/
/*************************************** Gui Thread *******************************************/
/***********************************************************************************************/
void *startGuiThread(void *)
{
	int i;
   int ans;
	int updtscrn = TRUE;

	/* delay to allow control thread to get up and running */
	while (GetMsg(TO_DISPLAY_MSG) != START_GUI) delay(50);

   while(terminate_gui == FALSE){

		/* update values */
		if (updtscrn == TRUE) UpdateScreen();

		ans = ' ';
		
		/* see if user has typed anything, execute if necessary */
		if((ans = getch()) != ERR){
			switch(ans){

				case ('q'): case ('Q'):
			   {	
					quitprogram();
				   break;
			   }
				case ('y'): case ('Y'):
			   {	
					if (updtscrn == TRUE) updtscrn = FALSE;
					else updtscrn = TRUE;
				   break;
			   }
				case KEY_UP:
				{
					SendMsg(TO_CONTROL_MSG,INCREASE_PARAMETER);
					break;
				}
				case KEY_DOWN:
				{
					SendMsg(TO_CONTROL_MSG,DECREASE_PARAMETER);
					break;
				}
				case KEY_LEFT:
				{
					SendMsg(TO_CONTROL_MSG,DECREASE_FIELD);
					break;
				}
				case KEY_RIGHT:
				{
					SendMsg(TO_CONTROL_MSG,INCREASE_FIELD);
					break;
				}
				case ('o'): case ('O'):
				{
					ScrollErrorMsg(-1);
					break;
				}
				case ('k'): case ('K'):
				{
					ScrollErrorMsg(1);
					break;
				}
				case ('d'): case ('D'):
				{
               /* redraw the whole screen if something got out of hand */
					werase(MainWindow);
					PrintErrorMsg("Redrawing Screen");
					DrawStaticInfo();
					break;
				}
				case ('t'): case ('T'):
				{
					SendMsg(TO_CONTROL_MSG,ENTER_TEST_MODE);
					break;
				}
				case ('r'): case('R'):
				{
					/* make sure the recording thread is still alive when you do this */
					if(Wait_for_RecordThread == TRUE){
						SendMsg(TO_RECORD_MSG,RESUME_RECORDING);
						sem_getvalue(&Data_Cache_Sem,&data_semval);
						if(data_semval == 0) sem_post(&Data_Cache_Sem);  /* need to post this to release the recording thread */
					}
					break;
				}
				case ('p'): case ('P'):
				{
					/* make sure the recording thread is still alive when you do this */
					if(Wait_for_RecordThread == TRUE){
						SendMsg(TO_RECORD_MSG,SUSPEND_RECORDING);
						sem_getvalue(&Data_Cache_Sem,&data_semval);
						if(data_semval == 0) sem_post(&Data_Cache_Sem);  /* need to post this to release the recording thread */
					}
					break;
				}
				case ('n'): case('N'):
				{
					/* make sure the recording thread is still alive when you do this */
					if(Wait_for_RecordThread == TRUE){
						SendMsg(TO_RECORD_MSG,START_NEW_FILE);
						sem_getvalue(&Data_Cache_Sem,&data_semval);
						if(data_semval == 0) sem_post(&Data_Cache_Sem);  /* need to post this to release the recording thread */
					}
					break;
				}
				default:
				{
				  PrintErrorMsg("User input not recognized...");
				}
			}  /* end of switch */
			  
			flushinp();
			  
		} /* if there's something in the buffer */
		
		else flushinp(); /* query user input */

		delay(DISPLAY_FREQUENCY);

	}// end infinite while loop

	PrintErrorMsg("TERMINATING GuiThread...");
	UpdateScreen();
   delay(500);

	Free_Windows();

   endwin();

	free(old_msgs);
   ThreadDestroy(0,-1,NULL);
   return(0);
}


void quitprogram(){

	PrintErrorMsg("Exiting...");
	SendMsg(TO_CONTROL_MSG,TERMINATE_CONTROL_THREAD);

}