/***************************************************************************/
/*                                                                         */
/* servotogo.c                                                             */
/* This file contains driver functions for the ServoToGo card              */
/* Created by ServoToGo, Inc.                                              */
/* Heavily Modified 11-22-2002 by Tom Worsnopp                             */
/* Minorly Modified 01-22-2003 by Eric Faulring                            */
/*                                                                         */
/* These driver functions work for the ISA Model 2 ServoToGo I/O Boards    */
/*                                                                         */
/***************************************************************************/

#define __STG
#include "servotogo.h"

/***************************************************************************/
/*                                                                         */
/*                               I/O                                       */
/*                                                                         */
/***************************************************************************/
inline void fOutP(unsigned short port, unsigned char data){
  out8(port,data);
}

inline void fOutPW(unsigned short port, unsigned short data){
  out16(port,data);
}

inline unsigned char fInP(unsigned short port){
  return in8(port);
}

inline unsigned short fInPW(unsigned short port){
  return in16(port);
}

/***************************************************************************/
/*                                                                         */
/*                           Initialize                                    */
/*                                                                         */
/***************************************************************************/
void stgInitialize(){
  // call this function to initialize the servotogo board

  unsigned short requested_irq = IRQ_NUMBER;

  mmap_device_io(32,0x300);         // get access to I/O ports under QNX, must be root
  ThreadCtl(_NTO_TCTL_IO,NULL);

  stgInitVariables();               // initialize global variables

  stgInitBaseAddress();             // determine board base address
  if (w_No_Board_Flag != NO_BOARD){
    if (w_Model == MODEL2){
      stgInitIRQ(requested_irq);    // initialize interrupt controller
      stgInitEncoders();
      stgInitDIOs();
      stgInitADCs();
      fOutP(stg_Base_Address + CNTRL1, CNTRL1_NOT_SLAVE);
      stgInitTimers();
    }
    else{
      //PrintErrorMsg('w',"The board you are using isn't a Model 2 Servotogo I/O Board.");
      exit(-1);
    }
  }
  else{
    //PrintErrorMsg('w',"The Servotogo Board was not detected.");
    exit(-1);
  }
}

/***************************************************************************/
/*                                                                         */
/*                            Disable Board                                */
/*                                                                         */
/***************************************************************************/
void stgDisable(){
  unsigned short i;
  stgStopInterrupts();
  
  // set all the DAC outputs to 0
  for (i = 0; i < MAX_AXIS; i++){
    stgWriteDAC(i, 0);
  }
  
  // set all the digital I/O bits to be inputs
  stgSetDioDirABC(ALL_INPUT);
  stgSetDioDirD(D_INPUT);
}

/***************************************************************************/
/*                                                                         */
/*                       Initialize Global Variables                       */
/*                                                                         */
/***************************************************************************/
void stgInitVariables(){
  int i;

  for (i = 0; i < MAX_AXIS; i++){
    by_Old_Byte2[i] = 0;
    by_Enc_High_Byte[i] = 0;
  }
  cps = SYSPAGE_ENTRY(qtime)->cycles_per_sec;
}

/***************************************************************************/
/*                                                                         */
/*                   Initialize Board Base Address                         */
/*                                                                         */
/***************************************************************************/
void stgInitBaseAddress(){
  stg_Base_Address = stgBaseFind();
  if (stg_Base_Address == 0){
    w_No_Board_Flag = NO_BOARD;
  }
  else{
    w_No_Board_Flag = BOARD_PRESENT;
  }
}

/***************************************************************************/
/*                                                                         */
/*                          Initialize IRQ                                 */
/*                                                                         */
/***************************************************************************/
void stgInitIRQ(unsigned short wRequestedIrq){
  // initialize the interrupt controller
  fOutP(stg_Base_Address + D_DIR, 0x82);  // initialize CNTRL0 as output reg
                                        // sets the direction of port D to output

  stgSetIRQ(wRequestedIrq);             // selects the IRQ in CNTRL0. Now, if a stray
                                        // interrupt is issued (see below) it will
                                        // go to an interrupt that isn't enabled on
                                        // the motherboard yet (if your system is
                                        // set up correctly).
}

/***************************************************************************/
/*                                                                         */
/*                         Initialize Encoders                             */
/*                                                                         */
/***************************************************************************/
void stgInitEncoders(){
  unsigned short    w_add;
  unsigned short    i;
  unsigned short    num_axes = 0;
  TYPE_LONGBYTE     enc[MAX_AXIS];

  // It is possible that the encoder counts are being held by battery
  // backup, so we'll read the encoders, and save the values
  // Then we'll initialize the encoder chips, since it's more likely that
  // the ecoders were not kept alive by battery and need to be initialized
  stgReadEncoders(enc);

  for (w_add = stg_Base_Address + CNT0_C;
       w_add <= stg_Base_Address + CNT6_C; w_add +=4){
    // we're going to be slick and do two chips at a time, that's why
    // the registers are arranged data, data, control, control.  You
    // can do two at a time, by using word operations, instead of
    // byte operations.  Not a big deal for initializing, but reading is
    // done pretty often.

    // Set Master Control Register
    fOutPW(w_add, 0x2020);       // reset master (MRST)

    // Set Input Control Register
    // Input Control, OL Load (P3), and Enable Inputs A and B (INA/B).
    fOutPW(w_add, 0x6868);

    // Set Output Control Register
    fOutPW(w_add, 0x8080);

    // Set Quadrature Register
    fOutPW(w_add, 0xc3c3);      // quadrature mode enabled

    fOutPW(w_add, 0x0404);      // reset counter to zero
  }

  // restore the values that you saved previously, in case encoder
  // was battery backed up

  for (w_add = stg_Base_Address + CNT0_D; w_add <= stg_Base_Address + CNT6_D; w_add += 4){

    // reset address pointer (RADR)
    fOutPW(w_add + 2, 0x0101);

    // place previous encoder values in preset register
    for (i = 0; i < 3; i++){
      fOutP(w_add, enc[num_axes].Byte[i]);
      fOutP(w_add + 1, enc[num_axes + 1].Byte[i]);
    }
    num_axes += 2;

    // transfer the preset register to the count register and reset address pointer
    fOutPW(w_add + 2, 0x0909);
  }
}

/***************************************************************************/
/*                                                                         */
/*                         Initialize ADCs                                 */
/*                                                                         */
/***************************************************************************/
void stgInitADCs(){
  /***********************************************************/
  /* SPECIAL NOTE ABOUT USING ADC AND DIGITAL I/O ON PORT D  */
  /*                                                         */
  /* Port D must be set to output to use the ADC.  Changing  */
  /* the direction of port D causes a calibration routine    */
  /* to run on the ADC.  This routine takes approximately    */
  /* 750 microseconds to complete.  The ADC is unuseable     */
  /* while the routine is running, so a delay must be added  */
  /* in your code before using the ADC.                      */
  /*                                                         */
  /* To avoid this problem altogether, don't use port D for  */
  /* input.  Then you'll never have to change the direction  */
  /* of port D.                                              */
  /***********************************************************/

  stgSetDioDirD(D_OUTPUT);              // set the direction of D port to output
  fOutP(stg_Base_Address + PORT_D, 0x00); // set ditial i/o on port D to be initially low
  stgWaitMicroSeconds(1000);            // delay 1 ms to let calibration routine finish
                                        // on ADC
}

/***************************************************************************/
/*                                                                         */
/*                    Initialize Digital I/O                               */
/*                                                                         */
/***************************************************************************/
void stgInitDIOs(){
  /***********************************************************/
  /* SPECIAL NOTE ABOUT USING ADC AND DIGITAL I/O ON PORT D  */
  /*                                                         */
  /* Port D must be set to output to use the ADC.  Changing  */
  /* the direction of port D causes a calibration routine    */
  /* to run on the ADC.  This routine takes approximately    */
  /* 750 microseconds to complete.  The ADC is unuseable     */
  /* while the routine is running, so a delay must be added  */
  /* in your code before using the ADC.                      */
  /*                                                         */
  /* To avoid this problem altogether, don't use port D for  */
  /* input.  Then you'll never have to change the direction  */
  /* of port D.                                              */
  /***********************************************************/

  stgSetDioDirABC(ALL_INPUT); // set the direction of ports A, B, and C to input
  stgSetDioDirD(D_OUTPUT);    // set the direction of port D to output
  stgWriteDIOs(0x00000000);   // set all digital i/o to be initially low
  stgWaitMicroSeconds(1000);  // delay 1 ms to let calibration routine finish
                              // on ADC
}

/***************************************************************************/
/*                                                                         */
/*                         Initialize Timers                               */
/*                                                                         */
/***************************************************************************/
void stgInitTimers(){
  // mode 5 - external strobe; doesn't run in this state
  stgSetTimer(TIMER_1, TMR_C1 | TMR_LMSB | TMR_M5 | TMR_BIN, 0x0001);

  // mode 5 - external strobe; doesn't run in this state
  stgSetTimer(TIMER_0, TMR_C0 | TMR_LMSB | TMR_M5 | TMR_BIN, 0x0001);

  // mode 5 - external strobe; doesn't run in this state
  stgSetTimer(TIMER_2, TMR_C2 | TMR_LMSB | TMR_M5 | TMR_BIN, 0x0001);
}

/***************************************************************************/
/*                                                                         */
/*                              Set Timer                                  */
/*                                                                         */
/***************************************************************************/
void stgSetTimer(unsigned short w_timer, unsigned char by_timer_cmd,
                 unsigned short w_timer_count){
  unsigned char *byte_ptr = (unsigned char *)&w_timer_count;

  fOutP(stg_Base_Address + TMRCMD, by_timer_cmd);

  switch(by_timer_cmd & TMR_LMSB){
    case (TMR_LMSB):
      fOutP(stg_Base_Address + w_timer, *byte_ptr++);
      fOutP(stg_Base_Address + w_timer, *byte_ptr);
      break;
    case (TMR_LSB):
      fOutP(stg_Base_Address + w_timer, *byte_ptr);
      break;
    case (TMR_MSB):
      *byte_ptr++;
      fOutP(stg_Base_Address + w_timer, *byte_ptr);
      break;
    case (TMR_CL):
      break;
  }
}

/***************************************************************************/
/*                                                                         */
/*                        ConfigureTimerTwoFrequency                       */
/*                         (Courtesy of Mark Salada)                       */
/***************************************************************************/
double stgConfigureTimerTwoFrequency(double TimerTwoFrequency)
{
   /* the input frequency is in units of Hz, and the return value is the
    * frequency actually implemented by the timers */
   /* WARNING: the timers start to fire immediately - you must have the ISR set before
    * calling this routine!!! */
   double requested_period;
   unsigned short cntr_val = 40;

   if (TimerTwoFrequency > 0.0)
   {
		requested_period = 1.0/TimerTwoFrequency;
		if(fmod(requested_period,TIMER12_RES) >= 0.5)
		{
			cntr_val = (unsigned short)ceil(requested_period/TIMER12_RES);
		}
		else
		{
			cntr_val = (unsigned short)floor(requested_period/TIMER12_RES);
		}
		/* Set TIMERCMD - TIMER_2   (load counter 2,  mode = square wave) */
		fOutP(stg_Base_Address + TMRCMD, TMR_C2 | TMR_LMSB | TMR_M3 | TMR_BIN);
		fOutP(stg_Base_Address + TIMER_2, (unsigned char)(cntr_val&0x00ff)); /* load LSB */
		fOutP(stg_Base_Address + TIMER_2, (unsigned char)((cntr_val&0xff00)>>8)); /* load MSB */
	}

   else cntr_val = 40; /* set to indicate error */

   return 1.0/(cntr_val*TIMER12_RES);
}
/***************************************************************************/
/*                                                                         */
/*                          ConfigureInterruptTimer                        */
/*                         (Courtesy of Mark Salada)                       */
/***************************************************************************/
double stgConfigureInterruptTimer(double InterruptFrequency)
{
   /* the input frequency is in units of Hz, and the return value is the
    * frequency actually implemented by the timers */
   /* WARNING: the timers start to fire immediately - you must have the ISR set before
    * calling this routine!!! */
   double requested_period;
   unsigned short cntr_val = 40;

   if (InterruptFrequency > 0.0)
   {
		requested_period = 1.0/InterruptFrequency;
		if(fmod(requested_period,TIMER0_RES) >= 0.5)
		{
			cntr_val = (unsigned short)ceil(requested_period/(TIMER0_RES));
		}
		else
		{
			cntr_val = (unsigned short)floor(requested_period/(TIMER0_RES));
		}

		/* Set TIMERCMD - TIMER_1   (load counter 1,  mode = square wave) */
		fOutP(stg_Base_Address + TMRCMD, TMR_C1 | TMR_LMSB | TMR_M3 | TMR_BIN);
		/* Load value into TIMER_1 */
		fOutP(stg_Base_Address + TIMER_1, (unsigned char)(TIMER1_VALUE&0x00ff)); /* load LSB */
		fOutP(stg_Base_Address + TIMER_1, (unsigned char)((TIMER1_VALUE&0xff00)>>8)); /* load MSB */

		/* Set TIMERCMD - TIMER_0     (load counter 0, mode = real time interrupt) */
		fOutP(stg_Base_Address + TMRCMD, TMR_C0 | TMR_LMSB | TMR_M2 | TMR_BIN);
		/* Load value into TIMER_0, from value calculated above */
		fOutP(stg_Base_Address + TIMER_0, (unsigned char)(cntr_val&0x00ff)); /* load LSB */
		fOutP(stg_Base_Address + TIMER_0, (unsigned char)((cntr_val&0xff00)>>8)); /* load MSB */
	}
	else cntr_val = 40; /* set to indicate error */

	return 1.0/(cntr_val*TIMER0_RES);
}

/***************************************************************************/
/*                                                                         */
/*                Set DIO Direction for Ports A, B, and C                  */
/*                                                                         */
/***************************************************************************/
void stgSetDioDirABC(unsigned char by_port_dir){
  // set direction of Digital I/O for ports A, B, and C - default is input

  unsigned long   l_current_data;
  unsigned char   by_dir;
  by_dir = 0x9b;                            // initially all ports input
  by_dir &= ~by_port_dir;

  l_current_data = stgReadDIOs();           // save state of DIO on all ports

  fOutP(stg_Base_Address + ABC_DIR, by_dir);  // set port directions

  stgWriteDIOs(l_current_data);             // restore state of DIO on all ports
}

/***************************************************************************/
/*                                                                         */
/*                     Set DIO Direction for Port D                        */
/*                                                                         */
/***************************************************************************/
void stgSetDioDirD(unsigned char by_port_dir){
  // set direction of Digital I/O for port D - default is input

  /***********************************************************/
  /* SPECIAL NOTE ABOUT USING ADC AND DIGITAL I/O ON PORT D  */
  /*                                                         */
  /* Port D must be set to output to use the ADC.  Changing  */
  /* the direction of port D causes a calibration routine    */
  /* to run on the ADC.  This routine takes approximately    */
  /* 750 microseconds to complete.  The ADC is unuseable     */
  /* while the routine is running, so a delay must be added  */
  /* in your code before using the ADC.                      */
  /*                                                         */
  /* To avoid this problem altogether, don't use port D for  */
  /* input.  Then you'll never have to change the direction  */
  /* of port D.                                              */
  /***********************************************************/

  unsigned long   l_current_data;
  unsigned char   by_dir;
  unsigned char   by_save_cntrl1, by_save_cntrl0;

  by_dir = 0x8b;                                  // initially port set to input
  by_dir &= ~by_port_dir;

  by_save_cntrl0 = fInP(stg_Base_Address + CNTRL0); // CNTRL0 needs to be saved, because
                                                  // D_DIR reinitializes the 8255 which
                                                  // implements the CNTRL0 register

  by_save_cntrl1 = fInP(stg_Base_Address + CNTRL1); // save for interrupt enables

  // don't reset any latches; put in slave state;
  // disable interrupts, so the glitch in CTRL0 doesn't
  // cause an interrupt on wrong irq
  fOutP(stg_Base_Address + CNTRL1, 0xf0);

  l_current_data = stgReadDIOs();                 // save state of DIO on all ports

  fOutP(stg_Base_Address + D_DIR, by_dir);          // set port D direction

  stgWriteDIOs(l_current_data);                   // restore state of DIO on all ports

  // restore CNTRL0, because it was re-initialized, which
  // lost any previous contents.
  fOutP(stg_Base_Address + CNTRL0, by_save_cntrl0);

  // re-enable interrupts, and restore slave state, don't
  // reset any latches. (1111xxxx)
  fOutP(stg_Base_Address + CNTRL1, (by_save_cntrl1 & 0x0f) | 0xf0);
}

/***************************************************************************/
/*                                                                         */
/*                             Base Find                                   */
/*                                                                         */
/***************************************************************************/
unsigned short stgBaseFind(){
  // find the base address of the ServoToGo board

  short i;
  unsigned short io_add;

  for (i = 15; i >= 0; i--){                    // search all possible addresses
    io_add = i * 0x20 + 0x200;
    if ( stgBrdtstOK(io_add) ){
      return io_add;
    }
  }
  return 0;
}

/***************************************************************************/
/*                                                                         */
/*                             BrdtstOK                                    */
/*                                                                         */
/***************************************************************************/
unsigned short stgBrdtstOK(unsigned short base_address){
  // check the board tests register, figure out board model

  unsigned short brdtst_address;
  unsigned short ser_seq, high_nibble;
  short j;

  brdtst_address = base_address + BRDTST;

  ser_seq = 0;

  for (j = 7; j >= 0; j--){
    high_nibble = fInP(brdtst_address) >> 4;
    if (high_nibble & 8){     // is SER set
      // shift bit to position specifed by Q2, Q1, Q0
      // which are the lower three bits.  Put bit in ser_seq.
      ser_seq |= 1 << (high_nibble & 7);
    }
  }
  if (ser_seq == 0x74){     // SER sequence is 01110100
    w_Model = MODEL2;
    return 1;               //true
  }
  else{
    w_Model = NO_MODEL;
    return 0;               // false
  }
}

/***************************************************************************/
/*                                                                         */
/*            Wait Microseconds  (1.65 us = 2600 cycles overhead)          */
/*                                                                         */
/***************************************************************************/
void stgWaitMicroSeconds(int microseconds){
  uint64_t  time_in;
  int       num_counts;

  num_counts = (int)((double)microseconds*(double)cps/1000000.0);

  time_in = ClockCycles();
  while(ClockCycles() < num_counts + time_in - STGWMS_OVERHEAD){};

}
	 
/***************************************************************************/
/*                                                                         */
/*                               Read ADC                                  */
/*                                                                         */
/***************************************************************************/
double stgReadADC(unsigned short w_axis){
  short          w_counts = 0;
  unsigned char  by_cntrl0;

  if (w_axis < MAX_AXIS){
    by_cntrl0 = fInP(stg_Base_Address + CNTRL0) & 0x07; // save irq

    by_cntrl0 |= (w_axis << 4);                     // shift bits to AD2, AD1, AD0
    by_cntrl0 |= NOAUTOZERO;                        // set bit 0x80 high, no autozero
    by_cntrl0 |= DONOTCALIBRATE_ADC;                // set bit 0x08 high, not calibrating

    fOutP(stg_Base_Address + CNTRL0, by_cntrl0);      // select ADC channel

    stgWaitMicroSeconds(4);                            // wait 4 uS for settling time
                                                    // on the multiplexer and ADC

    fOutPW(stg_Base_Address + ADC, 0);                // now start conversion

    stgWaitMicroSeconds(30);                           // wait for conversion to finish

    w_counts = fInPW(stg_Base_Address + ADC);
  
    if (w_counts & 0x1000){    // is sign bit negative?
      w_counts |= 0xf000;      // sign extend
    }
    else{
      w_counts &= 0xfff;       // make sure high order bits are zero.
    }
    return (double)w_counts*STG_VOLTS_COUNT;
  }
  else{
    return 0;
  }
}
/***************************************************************************/
/*                                                                         */
/*                      Multiple Read ADC                                  */
/*								                                                    */
/***************************************************************************/
void stgmReadADCs(unsigned char chan, double *values){
	short          w_count = 0;
   unsigned char  orig_by_cntrl0 = 0, by_cntrl0 = 0;
   int i;
	uint64_t			startmux;
   int num_aquired = 0;
   /* this function automatically updates a global reserve of the measurements,
    * called Analog_Inputs[8] */
   
   if (chan){
      
		/* grab current CNTRL0 register, zero out ADC selection, set noautozero and no cal */  
      orig_by_cntrl0 = (fInP(stg_Base_Address + CNTRL0) & 0x07) | AUTOZERO | CALIBRATE_ADC;
      
      /* loop through the mask */
      for(i=0;i<8;i++){
	 
			/* check to convert */
			if(chan & (1<<i)){
				by_cntrl0 = orig_by_cntrl0  | ((unsigned char)i << 4);	/* add in channel */
				fOutP(stg_Base_Address + CNTRL0, by_cntrl0);					/* select ADC channel */
				if (num_aquired > 0){
					startmux = ClockCycles();
					if ((fInP(stg_Base_Address + BRDTST) & EOC) > 0){			/* poll on the EOC bit  */
						stg_adc_timeout_errors+=1;										/* increment error and leave *values[i] where it is */
					} else {
						w_count = fInPW(stg_Base_Address + ADC);					/* keep in mind - this is a 12-bit conversion that apparently has a 13th sign bit... */  
						if (w_count & 0x1000){											/* is sign bit negative? */
							w_count |= 0xf000;											/* extend sign through uppermost nibble */
						} else {
							 w_count &= 0x0fff;											/* make sure high order bits are zero. */
						}
						values[num_aquired-1] = (double)w_count*STG_VOLTS_COUNT;				/* store counts in global variable */
					}	
					stgWaitMicroSeconds((int)(4.0-(double)(ClockCycles()-startmux)*1e06/cps));/* 4 uS multiplexer settle */
					fOutPW(stg_Base_Address + ADC, 0);								/* now start conversion */
					stgWaitMicroSeconds(30);               						/* "19" uS conversion settle */
					num_aquired += 1;
				}
				else {
					stgWaitMicroSeconds(4);
					fOutPW(stg_Base_Address + ADC, 0);								/* now start conversion */
					stgWaitMicroSeconds(30);               						/* "19" uS conversion settle */
					num_aquired += 1;
				}

				if (num_aquired == NUM_STG_ADC_CHANNELS)	{				 
					if ((fInP(stg_Base_Address + BRDTST) & EOC) > 0){			/* poll on the EOC bit  */
						stg_adc_timeout_errors+=1;										/* increment error and leave *values[i] where it is */
					} else {
						w_count = fInPW(stg_Base_Address + ADC);					/* keep in mind - this is a 12-bit conversion that apparently has a 13th sign bit... */  
						if (w_count & 0x1000){											/* is sign bit negative? */
							w_count |= 0xf000;											/* extend sign through uppermost nibble */
						} else {
							 w_count &= 0x0fff;											/* make sure high order bits are zero. */
						}
						values[num_aquired-1] = (double)w_count*STG_VOLTS_COUNT;				/* store counts in global variable */
					}			 
				}
			} /* checking chan to convert */

			///* check to convert */
			//if(chan & (1<<i)){
			//	by_cntrl0 = orig_by_cntrl0  | ((unsigned char)i << 4);	/* add in channel */
			//	fOutP(stg_Base_Address + CNTRL0, by_cntrl0);					/* select ADC channel */
			//	stgWaitMicroSeconds(4);               							/* 4 uS multiplexer settle */
			//	fOutPW(stg_Base_Address + ADC, 0);								/* now start conversion */
			//	stgWaitMicroSeconds(21);               						/* "19" uS conversion settle */
			//	 
			//	if ((fInP(stg_Base_Address + BRDTST) & EOC) > 0){			/* poll on the EOC bit  */
			//		stg_adc_timeout_errors+=1;										/* increment error and leave *values[i] where it is */
			//	} else {
			//		w_count = fInPW(stg_Base_Address + ADC);					/* keep in mind - this is a 12-bit conversion that apparently has a 13th sign bit... */  
			//		if (w_count & 0x1000){											/* is sign bit negative? */
			//			w_count |= 0xf000;											/* extend sign through uppermost nibble */
			//		} else {
			//			 w_count &= 0x0fff;											/* make sure high order bits are zero. */
			//		}
			//		values[i + 1] = (double)w_count*STG_VOLTS_COUNT;				/* store counts in global variable */
			//	}			 
			//
			//} /* checking chan to convert */
	 
		} /* looping through mask */
      
   }/* check for non-zero mask */
   
}
/***************************************************************************/
/*                                                                         */
/*                           Initialize Index Latches                                */
/*                                                                         */
/***************************************************************************/
void stgInitializeIndexLatches(){
  // initialize all the necessary hardware for the index latches

  // make sure that TIMER_0 is not active, because it will latch the
  // encoders if it is
  stgSetTimer(TIMER_0, TMR_C0 | TMR_LMSB | TMR_M5 | TMR_BIN, 0x0001);
  stgSetIDLEN(0XFF);  // enable the index pulse for all eight encoders
  stgSetSELDI(0x00);  // select index pulses (not external latches) on all channels 
  stgResetIDL();      // reset the latches
}

/***************************************************************************/
/*                                                                         */
/*                           Latch Encoders                                */
/*                                                                         */
/***************************************************************************/
void stgLatchEncoders(){
  unsigned short w_add;

  // you might have timer0 latch the encoders in hardware, but normally
  // timer0 won't be running and we need to latch them ourselves

  for (w_add = stg_Base_Address + CNT0_C; w_add <= stg_Base_Address + CNT6_C; w_add +=4){
    fOutPW(w_add, 0x0303);
  }
}

/***************************************************************************/
/*                                                                         */
/*                             Read Encoders                               */
/*                                                                         */
/***************************************************************************/
void stgReadEncoders(TYPE_LONGBYTE *lb_enc){
  unsigned short   w_add;
  unsigned short   i;
  unsigned short   num_axes = 0;
  TYPE_WORDBYTE    wb_transfer;

  // Disable interrupts here?  No, the timer will latch new data in the
  // hardware anyway.  Maybe we should stop the timer?  In an interrupt
  // service routine, you're synchronized with the timer; so the readings
  // will never change while you're reading them.  If you're polling, you
  // would first latch the encoder counts with the EncoderLatch() function.
  // But, the timer could latch the counts again, in the middle of the read.
  // A critical section will help in some extreme cases.

  // reset counter internal addr ptr to point to first byte
  for (w_add = stg_Base_Address + CNT0_C; w_add <= stg_Base_Address + CNT6_C; w_add +=4){
    fOutPW(w_add, 0x0101);
  }

  for (i = 0; i < 3; i++){            // 24 bits means get 3 bytes each
    wb_transfer.Word = fInPW(stg_Base_Address + CNT0_D);
      
    lb_enc[0].Byte[i] = wb_transfer.Byte.high;
    lb_enc[1].Byte[i] = wb_transfer.Byte.low;
      
    wb_transfer.Word = fInPW(stg_Base_Address + CNT2_D);
      
    lb_enc[2].Byte[i] = wb_transfer.Byte.high;
    lb_enc[3].Byte[i] = wb_transfer.Byte.low;
      
    wb_transfer.Word = fInPW(stg_Base_Address + CNT4_D);
      
    lb_enc[4].Byte[i] = wb_transfer.Byte.high;
    lb_enc[5].Byte[i] = wb_transfer.Byte.low;
      
    wb_transfer.Word = fInPW(stg_Base_Address + CNT6_D);
      
    lb_enc[6].Byte[i] = wb_transfer.Byte.high;
    lb_enc[7].Byte[i] = wb_transfer.Byte.low;
  }

  // maintain the high byte, to extend the counter to 32 bits

  // base decisions to increment or decrement the high byte
  // on the highest 2 bits of the 24 bit value.  To get the
  // highest 2 bits, use 0xc0 as a mask on byte [2] (the third
  // byte).

  for (i = 0; i < MAX_AXIS; i++){
    // check for -1 to 0 transition
    if (((by_Old_Byte2[i] & 0xc0) == 0xc0) // 11xxxxxx
        && ( (lb_enc[i].Byte[2] & 0xc0) == 0)){    // 00xxxxxx
      by_Enc_High_Byte[i]++;
    }

    // check for 0 to -1 transition
    if (((by_Old_Byte2[i] & 0xc0) == 0)    // 00xxxxxx
        && ((lb_enc[i].Byte[2] & 0xc0) == 0xc0)){ // 11xxxxxx
      by_Enc_High_Byte[i]--;
    }

    lb_enc[i].Byte[3] = by_Enc_High_Byte[i];
    by_Old_Byte2[i] = lb_enc[i].Byte[2];    // current byte 2 becomes old one
  }
}

/***************************************************************************/
/*                                                                         */
/*                             Read Encoder                                */
/*                                                                         */
/***************************************************************************/
long stgReadEncoder(unsigned short w_axis){
  TYPE_LONGBYTE   enc;
  unsigned short  i;
  unsigned short  axis = w_axis;

  enc.Long = 0;
  if (w_axis < MAX_AXIS){
    w_axis = (w_axis << 1) - (0x01 & w_axis);  // multiply by 2; subtract 1 if w_axis is odd

    fOutP(stg_Base_Address + CNT0_C + w_axis, 0x01);

    for (i = 0; i < 3; i++){
      enc.Byte[i] = fInP(stg_Base_Address + CNT0_D + w_axis);
    }

    // maintain the high byte, to extend the counter to 32 bits

    // base decisions to increment or decrement the high byte
    // on the highest 2 bits of the 24 bit value.  To get the
    // highest 2 bits, use 0xc0 as a mask on byte [2] (the third
    // byte).

    // check for -1 to 0 transition
    if (((by_Old_Byte2[axis] & 0xc0) == 0xc0) // 11xxxxxx
        && ( (enc.Byte[2] & 0xc0) == 0)){    // 00xxxxxx
      by_Enc_High_Byte[axis]++;
    }

    // check for 0 to -1 transition
    if (((by_Old_Byte2[axis] & 0xc0) == 0)    // 00xxxxxx
        && ((enc.Byte[2] & 0xc0) == 0xc0)){ // 11xxxxxx
      by_Enc_High_Byte[axis]--;
    }

    enc.Byte[3] = by_Enc_High_Byte[axis];
    by_Old_Byte2[axis] = enc.Byte[2];    // current byte 2 becomes old one

    return (enc.Long);
  }
  return 0L;
}

/***************************************************************************/
/*                                                                         */
/*                          Set Encoder Counts                             */
/*                                                                         */
/***************************************************************************/
void stgSetEncoderCounts(unsigned short w_axis, long l_counts){
  unsigned short w_add;
  char *ByteUnion = (char *)&l_counts;  // get pointer to l_counts, so you can take it apart byte by byte
  
  w_add = stg_Base_Address + CNT0_D;
  w_add += (w_axis & 0x6) << 1;        // shift to multiply by 2

  // pairs of data regs seperated by pairs
  // of control regs, skip over control.
  w_add += w_axis & 1;
  fOutP(w_add, ByteUnion[0]);
  fOutP(w_add, ByteUnion[1]);
  fOutP(w_add, ByteUnion[2]);
  
  // transfer the preset register to the count register
  fOutP(w_add + 2, 0x09);
  
  // extend the 24 bit counter to 32 bits
  by_Enc_High_Byte[w_axis] = ByteUnion[3];
  by_Old_Byte2[w_axis] = ByteUnion[2];
}

/***************************************************************************/
/*                                                                         */
/*                          Read Digital Inputs                            */
/*                                                                         */
/***************************************************************************/
unsigned long stgReadDIOs(){
  TYPE_IO32 x_in_bits;
  
  x_in_bits.port.A = fInP(stg_Base_Address + PORT_A);
  x_in_bits.port.B = fInP(stg_Base_Address + PORT_B);
  x_in_bits.port.C = fInP(stg_Base_Address + PORT_C);
  x_in_bits.port.D = fInP(stg_Base_Address + PORT_D);
  return (x_in_bits.all);
}

/***************************************************************************/
/*                                                                         */
/*                     Read One Port of Digital Inputs                     */
/*                                                                         */
/***************************************************************************/
unsigned char stgReadPortDIOs(unsigned short n_port){

  /**************************************************************/
  /* read one port of digital i/o                               */
  /* n_port must be equal to PORT_A, PORT_B, PORT_C, or PORT_D  */
  /**************************************************************/

  return fInP(stg_Base_Address + n_port);
}

/***************************************************************************/
/*                                                                         */
/*                          Read One Digital Input                         */
/*                                                                         */
/***************************************************************************/
unsigned char stgReadDIO(unsigned char by_bit_number, unsigned short n_port){

  /**************************************************************/
  /* read one bit on one port of digital i/o                    */
  /* n_port must be equal to PORT_A, PORT_B, PORT_C, or PORT_D  */
  /* by_bit_number specifies which bit on the port to be read   */
  /* by_bit_number should be a number from 0 to 7              */
  /**************************************************************/

  unsigned char  by_data = fInP(stg_Base_Address + n_port);

  by_data &= (1 << by_bit_number);
  by_data = (by_data >> by_bit_number);

  return by_data;
}

/***************************************************************************/
/*                                                                         */
/*                          Write Digital Outputs                          */
/*                                                                         */
/***************************************************************************/
void stgWriteDIOs(unsigned long l_out_bits){
  // write ALL digital i/o on all ports
  // l_out_bits = DDCCBBAA

  fOutP(stg_Base_Address + PORT_A, ((TYPE_IO32 *)&l_out_bits)->port.A);
  fOutP(stg_Base_Address + PORT_B, ((TYPE_IO32 *)&l_out_bits)->port.B);
  fOutP(stg_Base_Address + PORT_C, ((TYPE_IO32 *)&l_out_bits)->port.C);
  fOutP(stg_Base_Address + PORT_D, ((TYPE_IO32 *)&l_out_bits)->port.D);
}

/***************************************************************************/
/*                                                                         */
/*                          Write Digital Output                           */
/*                                                                         */
/***************************************************************************/
void stgWriteDIO(unsigned char by_bit_number, unsigned char by_set_0_or_1,
                 unsigned short n_port){
  /****************************************************************************/
  /* write ONE digital i/o on one port                                        */
  /* n_port must be equal to PORT_A, PORT_B, PORT_C, or PORT_D                */
  /* PORT_A, PORT_B, PORT_C, and PORT_D are constants defined in servotogo.h  */
  /* by_set_0_or_1 <- set 1 for high, set 0 for low                           */
  /* by_bit_number <- specifies which of the 8 bits on the port selected to   */
  /*                  change                                                  */
  /****************************************************************************/

  unsigned char by_data;

  by_data = fInP(stg_Base_Address + n_port);
  if (by_set_0_or_1 == 1){
    by_data |= 1 << by_bit_number;
  }
  else{
    by_data &= ~(1u << by_bit_number);
  }
  fOutP(stg_Base_Address + n_port, by_data);
}

/***************************************************************************/
/*                                                                         */
/*                                BaseAddress                              */
/*                                                                         */
/***************************************************************************/
unsigned short  stgGetBaseAddress(){
  return stg_Base_Address;
}

/***************************************************************************/
/*                                                                         */
/*                                GetModel                                 */
/*                                                                         */
/***************************************************************************/
unsigned short  stgGetModel(){
  return w_Model;
}

/***************************************************************************/
/*                                                                         */
/*                           StopInterrupts                                */
/*                                                                         */
/***************************************************************************/
void stgStopInterrupts(){
  // we want to save the state of the slave mode, and set the
  // high nibble bits high, so you don't reset any latches.
  // disable all interrupts, since only one can be enabled at
  // a time (currently).  If more than one was enabled, it's
  // an error (currently).
  // 1111x000, where x means not changed.
  fOutP(stg_Base_Address + CNTRL1, (fInP(stg_Base_Address + CNTRL1) & CNTRL1_NOT_SLAVE) | 0xf0);
}

/***************************************************************************/
/*                                                                         */
/*                            Write DAC                                    */
/*                                                                         */
/***************************************************************************/
void stgWriteDAC(unsigned short w_axis, long l_counts){
  if ( w_axis < MAX_AXIS){        // is axis within range?
    
    // input / output:
    //    l_counts (decimal) ... -l_counts ... +0x1000 ... volts
    //     0x1000   (4096)      0xfffff000           0       +10
    //          0                        0      0x1000         0
    // 0xfffff001  (-4095)           0xfff      0x1fff       -10

    // So, the domain might be different than you expected. I expected:
    //     0xf000 (-4096)  to  0xfff (4095), rather than
    //     0xf001 (-4095)  to 0x1000 (4096)

    // reverse slope so positive counts give positive voltage
    l_counts = - l_counts;

    // shift for DAC
    l_counts += 0x1000;

    if (l_counts > 0x1FFF){    // clamp + output
      l_counts = 0x1FFF;
    }
    if (l_counts < 0){         // clamp - output
      l_counts = 0;
    }

    fOutPW(stg_Base_Address + DAC_0 + (w_axis << 1), (unsigned short)l_counts);
  }
}

/***************************************************************************/
/*                                                                         */
/*                             Set IRQ                                     */
/*                                                                         */
/***************************************************************************/
void stgSetIRQ(unsigned short requested_irq){
  unsigned char by_int_reg = 0x00;

  by_int_reg |= NOAUTOZERO;         // set bit 0x80 high, no autozero
  by_int_reg |= DONOTCALIBRATE_ADC; // set bit 0x08 high, not calibrating

  switch (requested_irq){
    case 3:
      break;                     // add zero
    case 5:
      by_int_reg |= 4;
      break;
    case 7:
      by_int_reg |= 2;
      break;
    case 9:
      by_int_reg |= 6;
      break;
    case 10:
      by_int_reg |= 5;
      break;
    case 11:
      by_int_reg |= 7;
      break;
    case 12:
      by_int_reg |= 3;
      break;
    case 15:
      by_int_reg |= 1;
      break;
    default:                    // ERROR, requested irq not valid, use 5
      by_int_reg |= 4;          // There is no safe value, leaving zero
                                // here would select IRQ 3 which is worse
                                // than 5 because IRQ 3 is usually for COM 2
      break;
  }

  fOutP(stg_Base_Address + CNTRL0, by_int_reg);
}

/***************************************************************************/
/*                                                                         */
/*                             Reset IDL                                   */
/*                                                                         */
/***************************************************************************/
void stgResetIDL(){
  fOutP(stg_Base_Address + IDL, 0x00);
}

/***************************************************************************/
/*                                                                         */
/*                             Set IDLEN                                   */
/*                                                                         */
/***************************************************************************/
void stgSetIDLEN(unsigned char by_idlen){
  /***************************************************************/
  /* Please Note:                                                */
  /* If you enable the index pulse (i.e. set any of the bits in  */
  /* IDLEN to 1) for an encoder, and you plan to use hardware    */
  /* interrupts to latch that same encoder, you must remember to */
  /* set that bit of IDLEN back to 0 before using the hardware   */
  /* interrupts.  If you don't, and an index pulse has occurred, */
  /* then the hardware interrupts will not be able to latch the  */
  /* encoder.                                                    */
  /***************************************************************/

  fOutP(stg_Base_Address + IDLEN, by_idlen);
}

/***************************************************************************/
/*                                                                         */
/*                             Set SELDI                                   */
/*                                                                         */
/***************************************************************************/
void stgSetSELDI(unsigned char by_seldi){
  fOutP(stg_Base_Address + SELDI, by_seldi);
}

/***************************************************************************/
/*                                                                         */
/*                               Get IDL                                   */
/*                                                                         */
/***************************************************************************/
unsigned char stgGetIDL(){
  return fInP(stg_Base_Address + IDL);
}
